import json
from hashlib import sha256
from typing import Callable, List, Optional

import tiktoken


def num_tokens_from_text(text: str, encoding_name: str = "cl100k_base") -> int:
    """
    Returns the number of tokens in a string.

    Args:
        text: The text to count tokens in.
        encoding_name: The name of the token encoding to use. Defaults to "cl100k_base".

    Returns:
        The number of tokens in the string.
    """
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(text))
    return num_tokens


def truncate_text_by_tokens(
    text: str,
    max_tokens: int,
    encoding: tiktoken.core.Encoding,
) -> str:
    """
    Truncates a text to a maximum number of tokens.

    Args:
        text: The text to truncate.
        max_tokens: The maximum number of tokens to truncate the text to.
        encoding: The encoding to use.

    Returns:
        The truncated text.
    """
    tokens = encoding.encode(text)
    truncated_tokens = tokens[:max_tokens]
    truncated_text = encoding.decode(truncated_tokens)

    return truncated_text


def format_texts_as_json(texts: List[str]) -> str:
    """
    Formats a list of texts into a single string to be used as a user message.
    Each text is assigned an ID, starting from 0. The returned JSON format
    helps the model distinguish between different texts, at the cost of
    increasing the number of tokens used.

    The token overhead for a single text that doesn't require escaping characters
    is 12 tokens. Escaping characters like quotes increases the overhead.

    The format is a JSON list of dictionaries, where each dictionary has an
    "id" key and a "text" key. The "id" key is an integer, and the "text" key
    is a string. This array of maps structure is easiest to parse by GPT models
    and handles edge cases like newlines in the text.

    Args:
        texts: A list of texts to format.

    Returns:
        A formatted string that can be used as a user message.
    """

    if not isinstance(texts, list):
        raise ValueError("texts must be a list.")

    text_dicts = [
        {
            "id": i,
            "text": text,
        }
        for i, text in enumerate(texts)
    ]

    # json.dumps escapes characters like quotes, which increases the token overhead
    # OpenAI models understand non-ascii characters, so we can use ensure_ascii=False
    # to avoid escaping characters.
    return json.dumps(text_dicts, ensure_ascii=False)


def format_texts_with_spaces(texts: List[str]) -> str:
    """
    Simple formatter that joins texts with spaces.
    """
    return " ".join(texts)


def binpack_texts_in_order(
    texts: List[str],
    formatter_function: Callable[[List[str]], str],
    max_tokens_per_bin: int,
    max_texts_per_bin: Optional[int] = None,
    encoding_name: str = "cl100k_base",
    long_text_handling: str = "error",
) -> List[List[str]]:
    """
    Binpacks a list of texts into a list of lists of texts, such that each list of texts
    has a total number of tokens less than or equal to max_tokens_per_bin and each list of texts
    has a number of texts less than or equal to max_texts_per_bin.

    The binpacking uses a naive greedy algorithm that maintains the order of the texts.

    Args:
        texts: List of texts to binpack. Empty texts are accepted, counted as 0 tokens
            each and count against max_texts_per_bin.
        formatter_function: A function that takes a list of texts and returns a single
            text. Defaults to None, which means that the texts are joined with spaces.
            This function is used to include the overhead of the formatter function in
            the binpacking. It is not used to format the output. Make sure to use
            the same formatter function when formatting the output for the model.
        max_tokens_per_bin: The maximum number of tokens per bin of formatted texts.
            Leave some room for relative to the model's context size to account for the tokens in the
            system message, function call, and function return.
        max_texts_per_bin: The maximum number of texts per list of texts. Defaults to None, which
            means that there is no limit on the number of texts per list of texts.
        encoding_name: The name of the encoding to use. Defaults to "cl100k_base".
        long_text_handling: How to handle texts that are longer than max_tokens_per_bin. Defaults
            to "error", which means that an error is raised. Can also be set to
            "truncate", which means that the text is truncated to max_tokens_per_bin.
            It is possible that more tokens are truncated than absolutely necessary
            due to overhead of the formatter function caused by escaping characters.

    Returns:
        A list of lists of texts. The order of the texts is preserved.
    """

    if not isinstance(texts, list):
        raise ValueError("texts must be a list.")

    if not max_texts_per_bin:
        max_texts_per_bin = len(texts)

    if max_texts_per_bin < 1:
        raise ValueError(
            f"max_texts_per_bin must be at least 1, but got {max_texts_per_bin}"
        )

    encoding = tiktoken.get_encoding(encoding_name)

    # Binpack the texts
    # Initialize the first bin
    bins = []
    current_bin = []

    for i, text in enumerate(texts):
        if len(current_bin) == max_texts_per_bin:
            # Start a new bin
            bins.append(current_bin)
            current_bin = []

        # Calculate how many tokens would be in the current bin if we added the text
        bin_tokens_with_new_text = len(
            encoding.encode(formatter_function(current_bin + [text]))
        )

        if bin_tokens_with_new_text > max_tokens_per_bin:  # doesn't fit
            if len(current_bin) > 0:
                # Start a new bin
                bins.append(current_bin)
                current_bin = []

            # Check if the text fits in a bin by itself
            tokens_text_with_formatting = len(
                encoding.encode(formatter_function([text]))
            )

            if tokens_text_with_formatting > max_tokens_per_bin:  # doesn't fit
                # Calculate the overhead of the formatter function
                tokens_text_raw = len(encoding.encode(text))
                overhead = tokens_text_with_formatting - tokens_text_raw

                if overhead > max_tokens_per_bin:
                    raise ValueError(
                        f"""
                        The formatting function adds {overhead} overhead tokens,
                        which exceeds the maximum number of tokens ({max_tokens_per_bin}) permitted.
                        """
                    )

                if bin_tokens_with_new_text > max_tokens_per_bin:
                    # The formatted text is too long to fit in a bin
                    if long_text_handling == "error":
                        raise ValueError(
                            f"""
                            The text at index {i} has {tokens_text_with_formatting} tokens, which
                            is greater than the maximum number of tokens ({max_tokens_per_bin}).
                            Note that a formatting function added {overhead} tokens to the text.
                            """
                        )

                    elif long_text_handling == "truncate":
                        # Truncate the text, accounting for overhead
                        # It's possible that more is truncated than necessary
                        # in case the overhead was caused by escaping characters
                        # in the truncated part of the text
                        text = truncate_text_by_tokens(
                            text=text,
                            max_tokens=max_tokens_per_bin - overhead,
                            encoding=encoding,
                        )

                        assert (
                            len(encoding.encode(formatter_function([text])))
                            <= max_tokens_per_bin
                        )

                    else:
                        raise ValueError(
                            f"""
                            Invalid value for long_text_handling: {long_text_handling}.
                            Must be one of "error" or "truncate".
                            """
                        )

        # Add to the current bin
        current_bin.append(text)

    # Add the last bin
    bins.append(current_bin)

    return bins


def hash_dict(d: dict) -> str:
    """
    Hashes a dictionary using sha256.
    """
    return sha256(json.dumps(d).encode("utf-8")).hexdigest()



from dataclasses import dataclass


@dataclass
class Model:
    """
    Information about an OpenAI ChatCompletion model.
    Check prices here: https://openai.com/pricing

    Note that rate limits differ between OpenAI accounts.
    Check them here: https://platform.openai.com/account/rate-limits

    Args:
        name: The name of the model, e.g. "gpt-3.5-turbo".
        context_size: The maximum number of tokens that can be passed to the model.
        input_token_price_per_1k: The price in USD per 1000 tokens for input.
        output_token_price_per_1k: The price in USD per 1000 tokens for output.
        tokens_per_minute: The maximum number of tokens that can be processed per minute.
            Note that this may differ between OpenAI accounts. Override the default
            models' values with your own values.
        requests_per_minute: The maximum number of requests that can be made per minute.
            Note that this may differ between OpenAI accounts. Override the default
            models' values with your own values.
    """

    name: str
    context_size: int
    input_token_price_per_1k: float
    output_token_price_per_1k: float
    tokens_per_minute: int
    requests_per_minute: int

    def __post_init__(self):
        # Check that inputs are positive

        for arg in [
            "context_size",
            "input_token_price_per_1k",
            "output_token_price_per_1k",
            "tokens_per_minute",
            "requests_per_minute",
        ]:
            if getattr(self, arg) < 0:
                raise ValueError(f"{arg} must be positive")


# Look up information on models, pricing and rate limits:
# https://platform.openai.com/docs/models/overview
# https://openai.com/pricing
# https://platform.openai.com/account/rate-limits

GPT_4 = Model(
    name="gpt-4",
    context_size=8192,
    input_token_price_per_1k=0.03,
    output_token_price_per_1k=0.06,
    tokens_per_minute=10000,
    requests_per_minute=200,
)

GPT_4_0613 = Model(
    name="gpt-4-0613",
    context_size=8192,
    input_token_price_per_1k=0.03,
    output_token_price_per_1k=0.06,
    tokens_per_minute=10000,
    requests_per_minute=200,
)

GPT_4_32K = Model(
    name="gpt-4-32k",
    context_size=32768,
    input_token_price_per_1k=0.06,
    output_token_price_per_1k=0.12,
    tokens_per_minute=20000,
    requests_per_minute=200,
)

GPT_4_32K_0613 = Model(
    name="gpt-4-32k-0613",
    context_size=32768,
    input_token_price_per_1k=0.06,
    output_token_price_per_1k=0.12,
    tokens_per_minute=20000,
    requests_per_minute=200,
)

# legacy
GPT_4_0314 = Model(
    name="gpt-4-0314",
    context_size=8192,
    input_token_price_per_1k=0.03,
    output_token_price_per_1k=0.06,
    tokens_per_minute=10000,
    requests_per_minute=200,
)

# legacy
GPT_4_32K_0314 = Model(
    name="gpt-4-32k-0314",
    context_size=32768,
    input_token_price_per_1k=0.06,
    output_token_price_per_1k=0.12,
    tokens_per_minute=10000,
    requests_per_minute=200,
)

GPT_3_5_TURBO = Model(
    name="gpt-3.5-turbo",
    context_size=4096,
    input_token_price_per_1k=0.0015,
    output_token_price_per_1k=0.002,
    tokens_per_minute=90000,
    requests_per_minute=3500,
)

GPT_3_5_TURBO_16K = Model(
    name="gpt-3.5-turbo-16k",
    context_size=16384,
    input_token_price_per_1k=0.003,
    output_token_price_per_1k=0.004,
    tokens_per_minute=180000,
    requests_per_minute=3500,
)

GPT_3_5_TURBO_0613 = Model(
    name="gpt-3.5-turbo-0613",
    context_size=4096,
    input_token_price_per_1k=0.0015,
    output_token_price_per_1k=0.002,
    tokens_per_minute=90000,
    requests_per_minute=3500,
)

GPT_3_5_TURBO_16K_0613 = Model(
    name="gpt-3.5-turbo-16k-0613",
    context_size=16384,
    input_token_price_per_1k=0.003,
    output_token_price_per_1k=0.004,
    tokens_per_minute=180000,
    requests_per_minute=3500,
)

# legacy
GPT_3_5_TURBO_0301 = Model(
    name="gpt-3.5-turbo-0301",
    context_size=4096,
    input_token_price_per_1k=0.0015,
    output_token_price_per_1k=0.002,
    tokens_per_minute=9000,
    requests_per_minute=3500,
)


class Parameters:
    """
    Set of parameters that can be passed to an API request.

    The parameters are explained in the OpenAI API documentation:
    https://platform.openai.com/docs/api-reference/chat/create

    Args:
        max_tokens: The maximum number of tokens to generate. Note:
            This can't be greater than the model's context size and should be at least
            long enough to fit the whole expected JSON output. This parameter is used
            to estimate the cost of the request.
        temperature: What sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random, while
            lower values like 0.2 will make it more focused and deterministic.
            Defaults to 0.0 because this package is designed for deterministic
            JSON-schema compliant output.
        presence_penalty: Number between -2.0 and 2.0. Positive values penalize
            new tokens based on whether they appear in the text so far,
            increasing the model's likelihood to talk about new topics. Defaults to 0.0.
        frequency_penalty: Number between -2.0 and 2.0. Positive values penalize
            new tokens based on their existing frequency in the text so far,
            decreasing the model's likelihood to repeat the same line verbatim.
            Defaults to 0.0.

    Parameters that are not listed here are not supported by this package. The
    reason is that they're not relevant for the use case of this package.
    """

    def __init__(
        self,
        max_tokens: int,
        temperature: float = 0.0,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
    ):
        if max_tokens < 1:
            raise ValueError("max_tokens must be positive")

        if temperature < 0 or temperature > 1:
            raise ValueError("temperature must be between 0 and 1")

        if frequency_penalty < -2 or frequency_penalty > 2:
            raise ValueError("frequency_penalty must be between -2 and 2")

        if presence_penalty < -2 or presence_penalty > 2:
            raise ValueError("presence_penalty must be between -2 and 2")

        self.max_tokens = max_tokens
        self.temperature = temperature
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty

    def to_dict(self):
        """
        Returns:
            A dictionary representation of the parameters.
        """

        return {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
        }



# --------------------------------------------------------------------------------
# This file includes classes and functions adapted from: openai-cookbook
# Original source code: https://github.com/openai/openai-cookbook/blob/c651bfdda64ac049747c2a174cde1c946e2baf1d/examples/api_request_parallel_processor.py
# Copyright (c) 2023 OpenAI

# MIT License

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# imports
import asyncio  # for running API calls concurrently
import json  # for saving results to a jsonl file
import logging  # for logging rate limit warnings and other messages
import os  # for reading API key from environment variable
import sys  # for checking notebook vs. script
import tempfile  # for creating a temporary file to save results
import time  # for sleeping after rate limit is hit
from dataclasses import dataclass, field
from pathlib import Path  # for saving results to a file
from typing import Any, Dict, Generator, List, Optional, Union  # for type hints

import aiohttp
import aiohttp_client_cache

# for storing API inputs, outputs, and metadata
import jsonschema  # for validating API responses

from texttunnel.chat import ChatCompletionRequest
from texttunnel.models import Model
from texttunnel.utils import hash_dict

Response = List[Dict[str, Any]]

logger = logging.getLogger("texttunnel")


def prepare_output_filepath(
    output_filepath: Optional[Union[str, Path]], keep_file: bool
) -> Path:
    """
    Validates the output_filepath and returns a Path object. Uses a temporary file
    if output_filepath is None.

    Args:
        output_filepath: The path to save the results to. If None, a temporary file
            will be used.
        keep_file: Whether to keep the file after the function returns. If True,
            output_filepath must not be None.

    Returns:
        A Path object representing the output_filepath.
    """
    using_tempfile = False

    if output_filepath is None:
        output_filepath = tempfile.NamedTemporaryFile(delete=False).name
        using_tempfile = True
        if keep_file:
            raise ValueError(
                "keep_file=True is not compatible with output_filepath=None"
            )

    if not isinstance(output_filepath, Path):
        output_filepath = Path(output_filepath)

    if output_filepath.exists() and not using_tempfile:
        raise ValueError(f"File already exists: {output_filepath}")

    return output_filepath


def process_api_requests(
    requests: List[ChatCompletionRequest],
    output_filepath: Optional[Union[str, Path]] = None,
    keep_file: bool = False,
    max_attempts: int = 10,
    rate_limit_headroom_factor: float = 0.75,
    api_key: Optional[str] = None,
    cache: Optional[aiohttp_client_cache.CacheBackend] = None,
) -> List[Response]:
    """
    Make requests to OpenAI. This function is a wrapper around
    aprocess_api_requests() that executes it within asyncio.run, saving you the
    trouble of having to use asyncio directly.

    Note that if you're running this function in a Jupyter notebook, the function
    will automatically import nest_asyncio and call nest_asyncio.apply() to allow
    a second event loop to run in the same process. This is necessary because
    Jupyter notebooks already run an event loop in the background.

    If you require more control over the event loop, use the coroutine
    aprocess_api_requests() instead.

    Args:
        requests: List[ChatCompletionRequest]
            The requests to process, see ChatCompletionRequest class for details.
            Duplicate requests are not allowed.
        output_filepath: str, optional
            Path to the file where the results will be saved
            file will be a jsonl file, where each line is an array with the original
            request plus the API response e.g.,
            [{"model": "gpt-4", "messages": "..."}, {...}]
            if omitted, the results will be saved to a temporary file.
        keep_file: bool, optional
            Whether to keep the results file after the script finishes, in addition
            to the results being returned by the function.
            Defaults to False, so the file will be deleted after the script finishes.
            Setting this to True is not compatible with output_filepath=None.
        max_attempts: int, optional
            Number of times to retry a failed request before giving up
            if omitted, will default to 5
        rate_limit_headroom_factor: float, optional
            Factor to multiply the rate limit by to guarantee that the script
            stays under the limit if omitted, will default to 0.75
            (75% of the rate limit)
        api_key: str, optional
            API key to use. If omitted, the function will attempt to read it
            from an environment variable OPENAI_API_KEY. If that fails, an error
            will be raised, unless all requests are cached.
        cache: aiohttp_client_cache.CacheBackend, optional
            If provided, API responses will be served from the cache if available.
            New responses will be saved to the cache.
            Check the aiohttp_client_cache documentation for details on the
            available cache backends and how to configure them. See
            https://aiohttp-client-cache.readthedocs.io/en/stable/backends.html.
            Each backend requires different dependencies. For example, the SQLite
            backend requires the package "aiosqlite" to be installed.

    Returns:
        List[Dict[str, Any]]: list where each element consists of two dictionaries:
            - the original request
            - the API response
    """

    # Handle Notebook environment
    if "ipykernel" in sys.modules:
        # nest_asyncio is a workaround for running asyncio in Jupyter notebooks
        # it's always available when ipykernel is available
        import nest_asyncio

        nest_asyncio.apply()
        logger.info(
            "Running in Jupyter notebook environment. Activated nest_asyncio to allow asyncio to run."
        )

    responses = asyncio.run(
        aprocess_api_requests(
            requests=requests,
            output_filepath=output_filepath,
            keep_file=keep_file,
            max_attempts=max_attempts,
            rate_limit_headroom_factor=rate_limit_headroom_factor,
            api_key=api_key,
            cache=cache,
        )
    )

    return responses


async def fetch_json_response_from_cache(
    cache: aiohttp_client_cache.CacheBackend, url: str, request_json: dict
) -> Optional[dict]:
    """
    Fetch a response from the cache if it exists.

    Args:
        cache: Cache to fetch from.
        url: URL that was requested.
        request_json: JSON payload that was sent with the request.

    Returns:
        The cached response JSON if it exists, otherwise None.
    """
    cache_return_tuple = await cache.request(
        method="POST",  # ChatCompletion always uses POST requests
        url=url,
        json=request_json,
    )

    if cache_return_tuple[0] is None:
        return None

    cache_response_json = await cache_return_tuple[0].json()

    return cache_response_json


async def aprocess_api_requests(
    requests: List[ChatCompletionRequest],
    output_filepath: Optional[Union[str, Path]] = None,
    keep_file: bool = False,
    max_attempts: int = 10,
    rate_limit_headroom_factor: float = 0.75,
    api_key: Optional[str] = None,
    cache: Optional[aiohttp_client_cache.CacheBackend] = None,
) -> List[Response]:
    """
    Make asynchronous requests to the OpenAI API while
    throttling to stay under rate limits.

    Features:
    - Streams requests from file, to avoid running out of memory for giant jobs
    - Makes requests concurrently, to maximize throughput
    - Throttles request and token usage, to stay under rate limits
    - Retries failed requests up to {max_attempts} times, to avoid missing data
    - Logs errors, to diagnose problems with requests


    Args:
        requests: List[ChatCompletionRequest]
            The requests to process, see ChatCompletionRequest class for details.
            Duplicate requests are not allowed.
        output_filepath: str, optional
            Path to the file where the results will be saved
            file will be a jsonl file, where each line is an array with the original
            request plus the API response e.g.,
            [{"model": "gpt-4", "messages": "..."}, {...}]
            if omitted, the results will be saved to a temporary file.
        keep_file: bool, optional
            Whether to keep the results file after the script finishes, in addition
            to the results being returned by the function.
            Defaults to False, so the file will be deleted after the script finishes.
            Setting this to True is not compatible with output_filepath=None.
        max_attempts: int, optional
            Number of times to retry a failed request before giving up
            if omitted, defaults to 5.
        rate_limit_headroom_factor: float, optional
            Factor to multiply the rate limit by to guarantee that the script
            stays under the limit if omitted, defaults to 0.75.
            (75% of the rate limit).
        api_key: str, optional
            API key to use. If omitted, the function will attempt to read it
            from an environment variable OPENAI_API_KEY. If that fails, an error
            will be raised, unless all requests are cached.
        cache: aiohttp_client_cache.CacheBackend, optional
            If provided, API responses will be served from the cache if available.
            New responses will be saved to the cache.
            Check the aiohttp_client_cache documentation for details on the
            available cache backends and how to configure them. See
            https://aiohttp-client-cache.readthedocs.io/en/stable/backends.html.
            Each backend has different dependencies. For example, the SQLite
            backend requires the package "aiosqlite" to be installed.

    Returns:
        List[Dict[str, Any]]: list where each element consists of two dictionaries:
            - the original request
            - the API response
    """

    if len(requests) != len(set([request.get_hash() for request in requests])):
        # Duplicate requests can cause problems with ordering of results
        # Plus it's probably a mistake and would waste money
        raise ValueError("Duplicate requests detected. Each request must be unique.")

    # This function was adapted from openai-cookbook

    # The function is structured as follows:
    #    - Initialize things
    #    - In API processing loop
    #        - Get next request if one is not already waiting for capacity
    #        - Update available token & request capacity
    #        - If enough capacity available, call API
    #        - The loop pauses if a rate limit error is hit
    #        - The loop breaks when no tasks remain
    #   - Fetch results from file
    #   - Sort results in order of input requests
    #   - Return results

    output_filepath = prepare_output_filepath(output_filepath, keep_file)

    # Remember the order of the requests so that we can sort the results
    # Duplicate requests are not allowed, so the hash of each request is unique
    request_order = {request.get_hash(): i for i, request in enumerate(requests)}

    request_url = "https://api.openai.com/v1/chat/completions"

    if cache:
        check_cache_settings(cache)

        # Check if requests can be served from the cache
        # Build a queue of requests that need to be sent to the API
        # Handling cached requests separately allows us to avoid allocating
        # rate limit capacity to them and provide clearer logging.
        logger.debug("Checking cache for requests.")

        # Make asynchronous calls to the cache
        tasks = [
            fetch_json_response_from_cache(
                cache=cache,
                url=request_url,
                request_json=request.to_dict(),
            )
            for request in requests
        ]

        cached_responses = await asyncio.gather(*tasks)

        # Create list of requests that need to be sent to the API
        requests_queue = [
            request
            for request, response in zip(requests, cached_responses)
            if response is None
        ]

        # Write cached answers to file
        for request, response in zip(requests, cached_responses):
            if response is not None:
                data = [request.to_dict(), response]
                append_to_jsonl(data, output_filepath)

        request_cache_hits = len(requests) - len(requests_queue)
        logger.info(
            f"Found {request_cache_hits} out of {len(requests)} requests in cache."
        )
    else:
        logger.debug("No cache provided.")
        requests_queue = requests.copy()

    logger.debug("Cache check complete.")

    if len(requests_queue) > 0:
        await run_request_loop(
            requests_queue=requests_queue,
            request_url=request_url,
            output_filepath=output_filepath,
            cache=cache,
            max_attempts=max_attempts,
            rate_limit_headroom_factor=rate_limit_headroom_factor,
            api_key=api_key,
        )

    if cache:
        await cache.close()

    with open(output_filepath, "r") as f:
        request_response_pairs = [json.loads(line) for line in f]

    # Sort results in order of input requests
    # Results is a list of lists, where each sublist is [request, response]
    request_response_pairs = sorted(
        request_response_pairs,
        key=lambda x: request_order[hash_dict(x[0])],
    )

    assert len(request_response_pairs) == len(requests)

    if not keep_file:
        output_filepath.unlink()
    else:
        # Overwrite file with sorted results
        with open(output_filepath, "w") as f:
            for r in request_response_pairs:
                f.write(json.dumps(r) + "\n")

    return request_response_pairs


async def run_request_loop(
    requests_queue: List[ChatCompletionRequest],
    request_url: str,
    output_filepath: Path,
    cache: Optional[aiohttp_client_cache.CacheBackend] = None,
    max_attempts: int = 10,
    rate_limit_headroom_factor: float = 0.75,
    api_key: Optional[str] = None,
):
    """
    Run the main loop that processes API requests. Save results to a file.

    Args:
        requests_queue: A queue of requests to process.
        request_url: The URL to send the requests to.
        output_filepath: The path to the file where the results will be saved.
        cache: A aiohttp_client_cache.CacheBackend object that stores API
            responses. If provided, the response will be stored in the cache.
    """

    # Check that all requests use the same model. Otherwise, we can't set
    # a single rate limit for all requests.
    if len(set([request.model.name for request in requests_queue])) > 1:
        raise ValueError("All requests must use the same model.")

    if rate_limit_headroom_factor < 0.01 or rate_limit_headroom_factor > 1:
        raise ValueError("rate_limit_headroom_factor must be between 0.01 and 1.")

    # initialize API constants
    seconds_to_pause_after_rate_limit_error = 15
    seconds_to_sleep_each_loop = (
        0.001  # 1 ms limits max throughput to 1,000 requests per second
    )

    if api_key is None:
        api_key = fetch_api_key()

    request_header = {"Authorization": f"Bearer {api_key}"}

    # initialize trackers
    retry_queue = asyncio.Queue()
    task_id_generator = (
        task_id_generator_function()
    )  # generates integer IDs of 1, 2, 3, ...
    status_tracker = (
        StatusTracker()
    )  # single instance to track a collection of variables
    next_request = None  # variable to hold the next request to call

    # initialize available capacity counts
    max_requests_per_minute = (
        requests_queue[0].model.requests_per_minute * rate_limit_headroom_factor
    )
    max_tokens_per_minute = (
        requests_queue[0].model.tokens_per_minute * rate_limit_headroom_factor
    )

    available_request_capacity = max_requests_per_minute
    available_token_capacity = max_tokens_per_minute
    last_update_time = time.time()

    logger.debug("Initialization complete.")

    logger.info(
        f"Beginning main requests loop. {len(requests_queue)} requests to make."
    )

    # Main loop that runs until all tasks are finished
    last_status_log_timestamp = time.time()

    while True:
        # get next request (if one is not already waiting for capacity)
        # check if there are requests that need to be retried
        if next_request is None:
            if not retry_queue.empty():
                next_request = retry_queue.get_nowait()
                logger.debug(f"Retrying request {next_request.task_id}: {next_request}")
            # check if there are requests that haven't been tried yet
            if len(requests_queue) > 0:
                next_request = requests_queue.pop(0)

                # get new request
                next_request = APIRequest(
                    task_id=next(task_id_generator),
                    request=next_request,
                    token_consumption=next_request.count_total_tokens(),
                    attempts_left=max_attempts,
                )
                status_tracker.num_tasks_started += 1
                status_tracker.num_tasks_in_progress += 1
                logger.debug(f"Reading request {next_request.task_id}: {next_request}")

        # update available capacity
        current_time = time.time()
        seconds_since_update = current_time - last_update_time
        available_request_capacity = min(
            available_request_capacity
            + max_requests_per_minute * seconds_since_update / 60.0,
            max_requests_per_minute,
        )
        available_token_capacity = min(
            available_token_capacity
            + max_tokens_per_minute * seconds_since_update / 60.0,
            max_tokens_per_minute,
        )
        last_update_time = current_time

        # if enough capacity available, call API
        if next_request:
            next_request_tokens = next_request.token_consumption
            if (
                available_request_capacity >= 1
                and available_token_capacity >= next_request_tokens
            ):
                # update counters
                available_request_capacity -= 1
                available_token_capacity -= next_request_tokens
                next_request.attempts_left -= 1

                # call API
                asyncio.create_task(
                    next_request.call_api(
                        request_url=request_url,
                        request_header=request_header,
                        retry_queue=retry_queue,
                        output_filepath=output_filepath,
                        status_tracker=status_tracker,
                        cache=cache,
                    )
                )
                next_request = None  # reset next_request to empty

        # if all tasks are finished, break
        if status_tracker.num_tasks_in_progress == 0:
            break
        else:
            # Log status every 10 seconds
            if time.time() - last_status_log_timestamp > 10:
                logger.debug(
                    "%s tasks in progress. Successful tasks: %s. Failed tasks: %s. "
                    "Rate limit errors: %s. Other errors: %s. Retry queue length: %s ."
                    "Tasks not yet tried: %s. ",
                    status_tracker.num_tasks_in_progress,
                    status_tracker.num_tasks_succeeded,
                    status_tracker.num_tasks_failed,
                    status_tracker.num_rate_limit_errors,
                    status_tracker.num_other_errors,
                    retry_queue.qsize(),
                    len(requests_queue),
                )
                last_status_log_timestamp = time.time()

        # main loop sleeps briefly so concurrent tasks can run
        await asyncio.sleep(seconds_to_sleep_each_loop)

        # if a rate limit error was hit recently, pause to cool down
        seconds_since_rate_limit_error = (
            time.time() - status_tracker.time_of_last_rate_limit_error
        )
        if seconds_since_rate_limit_error < seconds_to_pause_after_rate_limit_error:
            remaining_seconds_to_pause = (
                seconds_to_pause_after_rate_limit_error - seconds_since_rate_limit_error
            )
            await asyncio.sleep(remaining_seconds_to_pause)
            # ^e.g., if pause is 15 seconds and final limit was hit 5 seconds ago
            logger.warn(
                f"Pausing to cool down until {time.ctime(status_tracker.time_of_last_rate_limit_error + seconds_to_pause_after_rate_limit_error)}"
            )

    # after finishing, log final status
    logger.info("Parallel processing complete.")
    if status_tracker.num_tasks_failed > 0:
        logger.warning(
            f"{status_tracker.num_tasks_failed} / {status_tracker.num_tasks_started} requests failed. Errors logged to {output_filepath}."
        )
    if status_tracker.num_rate_limit_errors > 0:
        logger.warning(
            f"{status_tracker.num_rate_limit_errors} rate limit errors received. Consider running at a lower rate."
        )


def check_cache_settings(cache: aiohttp_client_cache.CacheBackend) -> None:
    """
    Check that the cache is configured correctly to work with texttunnel.
    Raises a ValueError if the cache is not configured correctly.

    Args:
        cache: The cache to check.
    """
    if "POST" not in cache.allowed_methods:
        raise ValueError(
            'cache.allowed_methods must include "POST". Add the argument "allowed_methods=["POST"]" to the cache constructor.'
        )

    if cache.include_headers:
        raise ValueError("cache.include_headers must be False to protect the API key.")


def fetch_api_key() -> str:
    """
    Fetch the API key from the environment variable OPENAI_API_KEY. Raises a
    ValueError if the API key is not found.

    Returns:
        The API key.
    """

    try:
        api_key = os.getenv("OPENAI_API_KEY")
        assert api_key is not None
        return api_key
    except AssertionError:
        raise ValueError(
            "OPENAI_API_KEY environment variable not found. Please set it and try again."
        )


# dataclasses
@dataclass
class StatusTracker:
    """Stores metadata about the script's progress. Only one instance is created."""

    # This class was adapted from openai-cookbook

    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0  # script ends when this reaches 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0  # excluding rate limit errors, counted above
    num_other_errors: int = 0
    time_of_last_rate_limit_error: int = 0  # used to cool off after hitting rate limits


@dataclass
class APIRequest:
    """Stores an API request's inputs, outputs, and other metadata.
    Contains a method to make an API call."""

    task_id: int
    request: ChatCompletionRequest
    token_consumption: int
    attempts_left: int
    result: list = field(default_factory=list)

    # This class was adapted from openai-cookbook

    async def call_api(
        self,
        request_url: str,
        request_header: dict,
        retry_queue: asyncio.Queue,
        output_filepath: Path,
        status_tracker: StatusTracker,
        cache: Optional[aiohttp_client_cache.CacheBackend] = None,
        timeout_seconds: int = 120,
    ):
        """
        Calls the OpenAI API and appends the request and result to a JSONL file.
        If a cache provided, the result will be stored in the cache.
        The cache key is the hash of the request.

        Args:
            request_url: The URL to send the request to.
            request_header: The header to send with the request.
            retry_queue: A queue of requests that need to be retried.
                Will be populated if the request fails.
            output_filepath: The path to the file where the results will be saved.
            status_tracker: A StatusTracker object that tracks the greater
                request loop's progress.
            cache: A aiohttp_client_cache.CacheBackend object that stores API
                responses. If provided, the response will be stored in the cache.
            timeout_seconds: The number of seconds to wait for a response before
                timing out. Defaults to 120 seconds.
        """

        error = None

        logger.info(f"Starting request #{self.task_id}")
        timeout = aiohttp.ClientTimeout(total=timeout_seconds)

        # Choose the session class based on whether cache is provided
        session_class = (
            aiohttp.ClientSession
            if cache is None
            else aiohttp_client_cache.CachedSession
        )

        session_kwargs = (
            {"timeout": timeout}
            if cache is None
            else {"cache": cache, "timeout": timeout}
        )

        session = None

        try:
            session = session_class(**session_kwargs)  # type: ignore

            async with session.post(
                url=request_url, headers=request_header, json=self.request.to_dict()
            ) as response:
                response = await response.json()

            if "error" in response:
                logger.warning(
                    f"Request {self.task_id} failed with error {response['error']}"
                )
                status_tracker.num_api_errors += 1
                error = response
                if "Rate limit" in response["error"].get("message", ""):
                    status_tracker.time_of_last_rate_limit_error = int(time.time())
                    status_tracker.num_rate_limit_errors += 1
                    status_tracker.num_api_errors -= (
                        1  # rate limit errors are counted separately
                    )

        except (
            Exception
        ) as e:  # catching naked exceptions is bad practice, but in this case we'll log & save them
            logger.warning(f"Request {self.task_id} failed with Exception {e}")
            status_tracker.num_other_errors += 1
            error = e

        finally:
            # Always close the session - even if there was an error
            if session is not None:
                await session.close()

        if error:
            self.result.append(error)
            if self.attempts_left:
                retry_queue.put_nowait(self)
                logger.debug(
                    "Added request #%s to retry queue. Queue length: %s.",
                    self.task_id,
                    retry_queue.qsize(),
                )
            else:
                logger.error(
                    f"Request {self.request.to_dict()} failed after all attempts. Saving errors: {self.result}"
                )
                data = [self.request.to_dict(), [str(e) for e in self.result]]
                append_to_jsonl(data, output_filepath)
                status_tracker.num_tasks_in_progress -= 1
                status_tracker.num_tasks_failed += 1
        else:  # success
            data = [self.request.to_dict(), response]
            append_to_jsonl(data, output_filepath)
            status_tracker.num_tasks_in_progress -= 1
            status_tracker.num_tasks_succeeded += 1
            logger.debug(f"Request #{self.task_id} saved to {output_filepath}")


# functions
def append_to_jsonl(data: Any, filename: Path) -> None:
    """
    Append a json payload to the end of a jsonl file.

    Args:
        data: The data to append.
        filename: The file to append to.
    """
    # This function was adapted from openai-cookbook

    json_string = json.dumps(data)
    with open(filename, "a") as f:
        f.write(json_string + "\n")


def task_id_generator_function() -> Generator[int, None, None]:
    """
    Generate integers 0, 1, 2, and so on.

    Returns:
        A generator that yields integers.
    """
    # This function was adapted from openai-cookbook

    task_id = 0
    while True:
        yield task_id
        task_id += 1


RESPONSE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "array",
    "items": [
        # Request schema
        {
            "type": "object",
            "properties": {
                "model": {"type": "string"},
                "max_tokens": {"type": "integer"},
                "messages": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "role": {"type": "string"},
                            "content": {"type": "string"},
                        },
                        "required": ["role", "content"],
                    },
                },
                "functions": {
                    "type": "array",
                    "properties": {
                        "name": {"type": "string"},
                        "parameters": {"type": "object"},
                    },
                    "required": ["name", "parameters"],
                },
            },
            "required": [
                "model",
                "max_tokens",
                "messages",
                "functions",
            ],
        },
        # Response schema
        {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "object": {"type": "string"},
                "created": {"type": "integer"},
                "model": {"type": "string"},
                "choices": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "index": {"type": "integer"},
                            "message": {
                                "type": "object",
                                "properties": {
                                    "role": {"type": "string"},
                                    "content": {"type": ["string", "null"]},
                                    "function_call": {
                                        "type": "object",
                                        "properties": {
                                            "name": {"type": "string"},
                                            "arguments": {"type": "string"},
                                        },
                                        "required": ["name", "arguments"],
                                    },
                                },
                                "required": ["role", "function_call"],
                            },
                            "finish_reason": {"type": "string"},
                        },
                        "required": ["index", "message", "finish_reason"],
                    },
                },
                "usage": {
                    "type": "object",
                    "properties": {
                        "prompt_tokens": {"type": "integer"},
                        "completion_tokens": {"type": "integer"},
                        "total_tokens": {"type": "integer"},
                    },
                    "required": ["prompt_tokens", "completion_tokens", "total_tokens"],
                },
            },
            "required": ["id", "object", "created", "model", "choices", "usage"],
        },
    ],
}


def is_valid_response(response: Response, print_errors=False) -> bool:
    """
    Check if a response conforms to the response JSON schema.
    """
    try:
        jsonschema.validate(response, RESPONSE_SCHEMA)
        return True
    except jsonschema.exceptions.ValidationError as e:
        if print_errors:
            print(e)
        return False


def parse_arguments(response: Response) -> Dict[str, Any]:
    """
    Extract the function call arguments from a response.

    Args:
        response: The response to parse. It should be a list of length 2, where the
            first element is the request and the second element is the response.

    Returns:
        The function call arguments.
    """

    if not is_valid_response(response):
        raise ValueError("Response is not valid.")

    return json.loads(
        response[1]["choices"][0]["message"]["function_call"]["arguments"]
    )


def parse_token_usage(response: Response) -> Dict[str, Any]:
    """
    Extract the token usage from a response.

    Args:
        response: The response to parse. It should be a list of length 2, where the
            first element is the request and the second element is the response.

    Returns:
        The token usage.
    """
    if not is_valid_response(response):
        raise ValueError("Response is not valid.")

    return response[1]["usage"]


def usage_to_cost(usage: Dict, model: Model):
    """
    Convert token usage to cost in USD.

    Args:
        usage: The token usage. Retrieve it with parse_token_usage().
        model: The model used to generate the response.

    Returns:
        The cost in USD.
    """
    input_cost = model.input_token_price_per_1k * usage["prompt_tokens"] / 1000
    output_cost = model.output_token_price_per_1k * usage["completion_tokens"] / 1000
    total_cost = input_cost + output_cost
    return total_cost



# --------------------------------------------------------------------------------
# This file includes a function adapted from: openai-cookbook
# Original source code: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
# Copyright (c) 2023 OpenAI

# MIT License

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from typing import Any, Callable, Dict, List, Optional
from warnings import warn

import tiktoken
from jsonschema import Draft7Validator, exceptions

from texttunnel import utils
from texttunnel.models import Model, Parameters

FunctionDef = Dict[str, str]


def is_valid_function_def(function: FunctionDef) -> bool:
    """
    Checks if a function definition is valid for use in a ChatCompletionRequest.
    Note that the parameter properties are not validated to allow for custom properties.

    Check the OpenAI API documentation for more information:
    https://platform.openai.com/docs/guides/gpt/function-calling

    Args:
        function: The function definition to validate.
    """
    base_schema = {
        "name": {"type": "string"},
        "description": {"type": "string"},
        "parameters": {
            "type": "object",
            "properties": {"type": "object"},
        },
        "required": ["name", "parameters"],
    }

    try:
        Draft7Validator(base_schema).validate(function)
    except exceptions.ValidationError:
        print(f"Validation error: {exceptions.ValidationError}")
        return False

    return True


class ChatMessage:
    """
    A chat message, to be used in a chat.

    Args:
        role: The role of the message. Must be one of "system", "user", or "assistant".
        content: The content of the message.
    """

    VALID_ROLES = {"system", "user", "assistant"}

    def __init__(self, role: str, content: str):
        if role not in self.VALID_ROLES:
            raise ValueError(f"Invalid role {role}. Must be one of {self.VALID_ROLES}.")

        self.role = role
        self.content = content

    def to_dict(self) -> Dict[str, str]:
        """
        Returns a dict representation of the message.
        """
        return {"role": self.role, "content": self.content}


class Chat:
    """
    A chat. Used to prompt a model for a response.
    The first message must be from the system, and the last message must be from the user.

    Args:
        messages: A list of ChatMessage objects.
    """

    def __init__(self, messages: List[ChatMessage]):
        if len(messages) < 2:
            raise ValueError("A chat must have at least two messages.")

        if messages[0].role != "system":
            raise ValueError("The first message in a chat must be from the system.")

        if messages[-1].role != "user":
            raise ValueError("The last message in a chat must be from the user.")

        self.messages = messages

    def __len__(self) -> int:
        """
        Returns the number of messages in the chat.
        """
        return len(self.messages)

    def add_message(self, message: ChatMessage) -> None:
        """
        Adds a message to the end of the chat.

        Args:
            message: The message to add.
        """
        self.messages.append(message)

    def to_list(self) -> List[Dict[str, str]]:
        """
        Returns a list of dictionaries representing the chat messages.
        This is the format expected by the OpenAI API.
        """
        return [message.to_dict() for message in self.messages]

    def count_tokens(
        self,
        model: str = "gpt-3.5-turbo-0613",
        show_changing_model_warning: bool = False,
    ) -> int:
        """
        Return the number of tokens used.
        Note that this depends on the model used. Models that are not versioned
        with a date can change over time, causing an inaccurate token count
        by this function.

        Args:
            model: The name of the model to use. Defaults to "gpt-3.5-turbo-0613".
            show_changing_model_warning: Whether to print a warning if a model
                is used that may change over time. Defaults to False.

        Returns:
            The number of tokens used.
        """
        # This function was adapted from openai-cookbook

        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            if show_changing_model_warning:
                warn(
                    f"Warning: model {model} not found. Using cl100k_base encoding.",
                    UserWarning,
                )
            encoding = tiktoken.get_encoding("cl100k_base")

        # Map models that are not versioned with a date to a stable version
        stable_model_mapping = {
            "gpt-3.5-turbo": "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k": "gpt-3.5-turbo-16k-0613",
            "gpt-4": "gpt-4-0613",
            "gpt-4-32k": "gpt-4-32k-0613",
        }

        if model in stable_model_mapping:
            if show_changing_model_warning:
                warn(
                    f"Warning: {model} may update over time. Returning num tokens assuming {stable_model_mapping[model]}.",
                    UserWarning,
                )
            model = stable_model_mapping[model]

        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = (
                4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            )
            tokens_per_name = -1  # if there's a name, the role is omitted
        else:
            raise NotImplementedError(
                f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        num_tokens = 0
        for message in self.messages:
            num_tokens += tokens_per_message
            num_tokens += len(encoding.encode(message.content))
            if message.role == "name":
                num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens


class ChatCompletionRequest:
    """
    Defines a request for a chat completion.

    Args:
        chat: The chat to which the assistant should respond with a function call.
        model: The name of the OpenAI ChatCompletion model to use for completion.
        function: The function definition to use for the assistant's response.
            Must be a dictionary that describes a valid JSON schema.
            See https://platform.openai.com/docs/guides/gpt/function-calling
        params: Object of class Parameters. See models.Parameters for details.
    """

    def __init__(
        self,
        chat: Chat,
        model: Model,
        function: FunctionDef,
        params: Parameters,
    ):
        self.chat = chat
        self.model = model

        if not is_valid_function_def(function):
            raise ValueError("Invalid function definition.")

        self.function = function

        # Force the model to use a function call
        self.functions = [function]
        self.function_call = {"name": function["name"]}

        if params.max_tokens > self.model.context_size:
            raise ValueError(
                f"""
                max_tokens ({params.max_tokens}) exceeds the context
                size of the model ({self.model.context_size}).
                """
            )

        self.params = params

        # Check that the inputs fit into the context size and leaves
        # enough space for the output
        num_prompt_tokens = self.count_prompt_tokens()
        num_completion_tokens = self.count_completion_tokens()
        num_total_tokens = num_prompt_tokens + num_completion_tokens

        if num_total_tokens > self.model.context_size:
            raise ValueError(
                f"""
                Total number of tokens ({num_total_tokens}) exceeds the context
                size of the model ({self.model.context_size}). Input tokens:
                {num_prompt_tokens}. Output tokens: {num_completion_tokens}.
                Context size: {self.model.context_size}.
                """
            )

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns a dictionary representation of the request. Only includes
        the elements that are required by the OpenAI API. Model parameters
        are flattened into the top-level dictionary.
        """
        return {
            "model": self.model.name,
            "messages": self.chat.to_list(),
            "functions": self.functions,
            "function_call": self.function_call,
            **self.params.to_dict(),
        }

    def get_hash(self) -> str:
        """
        Returns the hash of the request. Can be used as a cache key.
        """
        return utils.hash_dict(self.to_dict())

    def count_prompt_tokens(self) -> int:
        """
        Counts the number of tokens that will be used as input to the model.
        This includes the chat messages and the function call.
        """
        chat_tokens = self.chat.count_tokens(model=self.model.name)
        function_tokens = utils.num_tokens_from_text(json.dumps(self.functions[0]))

        return chat_tokens + function_tokens

    def count_completion_tokens(self) -> int:
        """
        Counts the number of tokens that will be used as output of the model.
        Assumes that the model will return the maximum number of tokens allowed
        by the max_tokens parameter.
        """

        return self.params.max_tokens

    def count_total_tokens(self) -> int:
        """
        Counts the total number of tokens that will be used as input and output
        of the model. Assumes that the model will return the maximum number of
        tokens allowed by the max_tokens parameter.
        """
        return self.count_prompt_tokens() + self.count_completion_tokens()

    def estimate_cost_usd(self) -> float:
        """
        Estimates the cost of the request in USD. Assumes that the model will
        return the maximum number of tokens allowed by the max_tokens parameter.
        The estimate is the upper bound on the cost, since the model may return
        fewer tokens than the maximum allowed.
        """

        input_cost_usd = (
            self.count_prompt_tokens() * self.model.input_token_price_per_1k / 1000
        )
        output_cost_usd = (
            self.count_completion_tokens() * self.model.output_token_price_per_1k / 1000
        )

        return input_cost_usd + output_cost_usd


def build_binpacked_requests(
    model: Model,
    function: FunctionDef,
    system_message: str,
    texts: List[str],
    params: Parameters,
    max_tokens_per_request: Optional[int] = None,
    max_texts_per_request: Optional[int] = None,
    binpacking_function: Callable = utils.binpack_texts_in_order,
    formatter_function: Callable = utils.format_texts_as_json,
    encoding_name: str = "cl100k_base",
    long_text_handling: str = "error",
) -> List[ChatCompletionRequest]:
    """
    Builds a list of ChatCompletionRequests from a list of texts.
    If possible, multiple texts will be combined into a single ChatCompletionRequest.
    This can reduce the number of tokens spent on overheads like the system message
    and function definition.

    The requests can then be passed to processor.process_api_requests().

    Args:
        model: The model to use for completion.
        function: The function definition to use for the assistant's response.
            Must be a dictionary that describes a valid JSON schema.
            See https://platform.openai.com/docs/guides/gpt/function-calling
        system_message: The message to include at the beginning of each chat.
        texts: A list of texts to binpack into chats. Duplicates are not allowed.
        params: Object of class Parameters. See models.Parameters for details.
        max_tokens_per_request: The maximum number of tokens allowed in one request.
            Defaults to 90% of the model's context size. The 10% buffer makes
            sure that mistakes in token counting don't cause the request to fail.
        max_texts_per_request: The maximum number of texts allowed in one request.
            Defaults to None, which means there is no limit.
        binpacking_function: The function to use for binpacking.
            Must take a list of texts and return a list of lists of texts.
            Defaults to binpack_texts_in_order().
        formatter_function: The function to use for formatting the texts.
            Must take a list of texts and return a single string.
            Defaults to format_texts_as_json().
        encoding_name: The name of the encoding to use for tokenization.
            Defaults to "cl100k_base".
        long_text_handling: Passed to the binpacking function. Defaults to
            "error", which means that an error will be raised if a text is too
            long to fit in a single chat.

    Returns:
        A list of ChatCompletionRequests.
    """
    if len(set(texts)) != len(texts):
        # Downstream code assumes that each request has a unique hash
        # Duplicate texts would cause the requests to have the same hash
        # Plus it's probably a mistake and would waste money
        raise ValueError("Duplicate texts found. Please remove duplicates.")

    if max_tokens_per_request is None:
        max_tokens_per_request = int(model.context_size * 0.9)

    # System message and function definition count towards the token limit
    overheads = [system_message, json.dumps(function)]
    static_tokens = sum([utils.num_tokens_from_text(text) for text in overheads])

    # Calculate the maximum number of tokens left for the chat,
    # after accounting for the overheads and the output tokens
    max_tokens_per_chat = max_tokens_per_request - static_tokens - params.max_tokens

    # Binpack the texts into chats
    text_bins = binpacking_function(
        texts=texts,
        max_tokens_per_bin=max_tokens_per_chat,
        max_texts_per_bin=max_texts_per_request,
        encoding_name=encoding_name,
        formatter_function=formatter_function,
        long_text_handling=long_text_handling,
    )

    requests = []

    for text_bin in text_bins:
        # Create a chat from the bin
        messages = [ChatMessage("system", system_message)]
        messages.append(ChatMessage("user", formatter_function(text_bin)))

        chat = Chat(messages)

        request = ChatCompletionRequest(
            chat=chat,
            model=model,
            function=function,
            params=params,
        )

        requests.append(request)

    return requests


def build_requests(
    model: Model,
    function: FunctionDef,
    system_message: str,
    texts: List[str],
    params: Parameters,
    encoding_name: str = "cl100k_base",
    long_text_handling: str = "error",
) -> List[ChatCompletionRequest]:
    """
    Builds a list of ChatCompletionRequests from a list of texts.
    The requests can then be passed to processor.process_api_requests().

    Args:
        model: The model to use for completion.
        function: The function definition to use for the assistant's response.
            Must be a dictionary that describes a valid JSON schema.
            See https://platform.openai.com/docs/guides/gpt/function-calling
        system_message: The message to include at the beginning of each chat.
        params: Object of class Parameters. See models.Parameters for details.
        texts: A list of texts to binpack into chats. Duplicates are not allowed.
        encoding_name: The name of the encoding to use for tokenization.
            Defaults to "cl100k_base".
        long_text_handling: Passed to the binpacking function. Defaults to
            "error", which means that an error will be raised if a text is too
            long to fit in a single chat.

    Returns:
        A list of ChatCompletionRequests.
    """

    return build_binpacked_requests(
        model=model,
        function=function,
        system_message=system_message,
        texts=texts,
        params=params,
        max_tokens_per_request=None,
        max_texts_per_request=1,
        binpacking_function=utils.binpack_texts_in_order,
        formatter_function=utils.format_texts_with_spaces,
        encoding_name=encoding_name,
        long_text_handling=long_text_handling,
    )
